
class UnixPackageException(Exception):
    """Base class for all unixpackage exceptions."""
    pass


class RequirementsFileCannotBeRead(UnixPackageException):
    def __init__(self, filename):
        super(RequirementsFileCannotBeRead, self).__init__((
            "Error reading requirements file '{0}'\n"
        ).format(filename))


class PackageNotFoundInEquivalents(UnixPackageException):
    def __init__(self, package_equivalents, distro):
        super(PackageNotFoundInEquivalents, self).__init__((
            "Package for distro {1} not found in {0}\n"
            "Try running 'unixpackage cleancache' or "
            "raise an issue at https://github.com/unixpackage/unixpackage.github.io/"
        ).format(package_equivalents, distro))


class UnsupportedPlatform(UnixPackageException):
    def __init__(self, this_distro):
        super(UnsupportedPlatform, self).__init__((
            "Platform {0} is not supported yet.\n"
            "Raise an issue at http://github.com/unixpackage/unixpackage"
            " for help."
        ).format(this_distro))


class NoPackagesSpecified(UnixPackageException):
    def __init__(self):
        super(NoPackagesSpecified, self).__init__("No packages specified.")


class PackageNotFound(UnixPackageException):
    def __init__(self, generic_package_name, url):
        super(PackageNotFound, self).__init__((
            "Package {0} not found at {1}\n"
            "Please consider adding it if you think it should exist!\n"
            "Simply raise an issue or issue a pull request on: "
            "http://github.com/unixpackage/unixpackage.github.io"
        ).format(generic_package_name, url))


class NetworkError(UnixPackageException):
    def __init__(self, url, error):
        super(NetworkError, self).__init__((
            "Error querying https://unixpackage.github.io/{0}.json"
            "- Reason: {1}."
        ).format(package, error))


class PackageDescriptionNotUnderstood(UnixPackageException):
    def __init__(self, package):
        super(PackageDescriptionNotUnderstood, self).__init__((
            "Problem decoding package {0}. Please raise an issue at "
            "http://github.com/unixpackage/unixpackage.github.io"
        ).format(package))


class PackageInstallationFailed(UnixPackageException):
    def __init__(self, generic_package_names, installation_command):
        super(PackageInstallationFailed, self).__init__((
            "Package installation of {0} failed. Try again with: '{1}'"
        ).format(', '.join(generic_package_names), installation_command))


class ConnectionFailure(UnixPackageException):
    def __init__(self, url):
        super(ConnectionFailure, self).__init__((
                "Failure when connecting to {0}\n"
                "Is your internet working?"
            ).format(url))



class CalledProcessError(UnixPackageException):
    pass



from unixpackage.package_group import package_group_for_my_distro
from unixpackage.utils import log, warn, check_call
from unixpackage import exceptions
import signal
import os


def parse_requirements_file(filename):
    """Parse Requirements file and return a list of packages."""
    if filename is not None:
        filename = os.path.abspath(filename)
        try:
            with open(filename, "r") as req_file_handle:
                list_of_requirements = [
                    line for line in req_file_handle.read().split('\n') \
                        if line.strip() != "" and not line.startswith("#")
                ]
            return list_of_requirements
        except IOError:
            raise exceptions.RequirementsFileCannotBeRead(filename)
    else:
        return []


def install_command(generic_packages):
    """Return an install command for this distro."""
    if len(generic_packages) == 0:
        raise exceptions.NoPackagesSpecified()
    return package_group_for_my_distro()(generic_packages).install_cmd()


def packages_installed(generic_packages):
    """Verify that the list of packages is installed."""
    if len(generic_packages) == 0:
        raise exceptions.NoPackagesSpecified()
    if len(generic_packages) > 10:
        packages_installed(generic_packages[10:])
        generic_packages = generic_packages[:10]
    return package_group_for_my_distro()(generic_packages).check()


def install(generic_packages, polite=False):
    """Attempt installation of specified packages (if not already installed)."""
    if len(generic_packages) == 0:
        raise exceptions.NoPackagesSpecified()
    if len(generic_packages) > 10:
        install(generic_packages[10:])
        generic_packages = generic_packages[:10]

    package_group = package_group_for_my_distro()(generic_packages)

    if not package_group.check():
        not_preinstalled = package_group.not_installed()
        install_cmd = not_preinstalled.install_cmd()
        if package_group.need_sudo and polite:
            previous_signal = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            log((
                "The following command must be run to continue. I am attempting to run it now:"
                "\n\n\n       {0}\n\n\n"
                "You can also run this command in another window "
                "and then hit Ctrl-C to continue.\n\n"
            ).format(install_cmd))
        try:
            check_call(install_cmd, shell=True)
        except exceptions.CalledProcessError:
            warn("\nWARNING : Command '{0}' returned error code\n\n".format(install_cmd))
        if package_group.need_sudo and polite:
            signal.signal(signal.SIGINT, previous_signal)

        # Double check that the packages were correctly installed
        not_installed_after_install_cmd = not_preinstalled.not_installed()

        if not not_installed_after_install_cmd.empty():
            # Throw meaningful error with the command to re-run to install packages
            raise exceptions.PackageInstallationFailed(
                not_installed_after_install_cmd.generic_package_list,
                not_installed_after_install_cmd.install_cmd()
            )
        else:
            log("Post-install package check for {0} successful!\n".format(
                ", ".join(not_preinstalled.generic_package_list)
            ))
    else:
        log("Already installed: {0}\n".format(', '.join(generic_packages)))



from unixpackage import exceptions
from unixpackage import utils
from sys import stdout
from os import path
import platform
import json
import sys
import os


class PackageGroup(object):
    """Representation of a group of generic packages from generic package names."""
    need_sudo = True

    def __init__(self, generic_package_list):
        """Initialize and download/retrieve from cache distro specific package names."""
        self.generic_package_list = generic_package_list
        self.specific_packages = []

        for package in generic_package_list:
            cache_dir = path.join(path.expanduser("~"), ".unixpackage")
            if not path.exists(cache_dir):
                os.makedirs(cache_dir)

            cache_filename = path.join(cache_dir, "{0}.json".format(package))
            if path.exists(cache_filename):
                package_equivalents = utils.get_json_from_file(cache_filename)
            else:
                utils.log((
                    "Downloading and caching correct package names for: {0}...\n"
                ).format(package))

                url = "https://unixpackage.github.io/{0}.json".format(package)
                contents = utils.get_request(url)

                if contents is None:
                    raise exceptions.PackageNotFound(package, url)
                else:
                    package_equivalents = json.loads(contents)
                    utils.save_json_to_file(cache_filename, package_equivalents)

            specific_package_equivalent = self.get_specific_package(package_equivalents)

            if specific_package_equivalent is not None:
                if type(specific_package_equivalent) is list:
                    self.specific_packages.extend(specific_package_equivalent)
                elif utils.is_string(specific_package_equivalent):
                    self.specific_packages.append(specific_package_equivalent)
                elif specific_package_equivalent is None:
                    pass
                else:
                    raise PackageDescriptionNotUnderstood(package)

    def empty(self):
        """If the generic packages have no equivalent in this distro, this may be true."""
        return len(self.specific_packages) == 0

    def install_cmd(self):
        """Get the distro-specific install command for these packages."""
        return "{0}{1} {2}".format(
            "sudo " if self.need_sudo else "",
            self.install_prefix,
            " ".join(self.specific_packages),
        )

    def check(self):
        """Check if the packages are installed."""
        if self.empty():
            return True
        else:
            return utils.return_code_zero(self.check_cmd + self.specific_packages)

    def not_installed(self):
        """Return a package group of packages not installed from this group."""
        not_installed_list = []
        for package in self.generic_package_list:
            package_group_of_one = package_group_for_my_distro()([package])

            if not package_group_of_one.check():
                not_installed_list.append(package)
        return package_group_for_my_distro()(not_installed_list)

    def get_specific_package(self, package_equivalents):
        """Base method to get a specific package."""
        for x in self.__class__.__mro__[:-2]:
            if x.name in package_equivalents:
                return package_equivalents[x.name]
        raise exceptions.PackageNotFoundInEquivalents(package_equivalents, self.distro)


class DebPackageGroup(PackageGroup):
    need_sudo = True
    install_prefix = "apt-get install -y"
    check_cmd = ["dpkg", "--status", ]
    distro = "Debian"
    name = "basedeb"


class DebianPackageGroup(DebPackageGroup):
    name = "debian"


class DebianSqueezePackageGroup(DebianPackageGroup):
    name = "debian-squeeze"


class DebianSqueezeLTSPackageGroup(DebianPackageGroup):
    name = "debian-squeeze-lts"


class DebianWheezyPackageGroup(DebianPackageGroup):
    name = "debian-wheezy"


class DebianJessiePackageGroup(DebianPackageGroup):
    name = "debian-jessie"


class DebianStretchPackageGroup(DebianPackageGroup):
    name = "debian-stretch"


class DebianSidPackageGroup(DebianPackageGroup):
    name = "debian-sid"


class UbuntuPackageGroup(DebPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu"


class UbuntuPrecisePackageGroup(UbuntuPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu-precise"


class UbuntuTrustyPackageGroup(UbuntuPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu-trusty"


class UbuntuVividPackageGroup(UbuntuPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu-vivid"


class UbuntuWilyPackageGroup(UbuntuPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu-wily"


class UbuntuXenialPackageGroup(UbuntuPackageGroup):
    distro = "Ubuntu"
    name = "ubuntu-xenial"


class ArchPackageGroup(PackageGroup):
    need_sudo = True
    install_prefix = "pacman -Sy"
    check_cmd = ["pacman", "-Q", ]
    distro = "Arch"
    name = "arch"


class RPMPackageGroup(PackageGroup):
    need_sudo = True
    install_prefix = "yum -y install"
    check_cmd = ["rpm", "-q", ]
    distro = "RPM Distro"
    name = "baserpm"


class CentOSPackageGroup(RPMPackageGroup):
    distro = "CentOS"
    name = "centos"


class FedoraPackageGroup(RPMPackageGroup):
    distro = "Fedora"
    name = "fedora"


class RedHatPackageGroup(RPMPackageGroup):
    distro = "Red Hat"
    name = "redhat"


class MacOSBrewPackageGroup(PackageGroup):
    need_sudo = False
    install_prefix = "brew install"
    distro = "Mac OS X"
    name = "macosbrew"

    def check(self):
        """brew list --versions will simply not output not-installed packages specified."""
        if not self.empty():
            brew_list_output = utils.check_output(
                ["brew", "list", "--versions", ] + self.specific_packages
            ).decode("utf8").rstrip()
            brew_list_installed_pkgs = [x for x in brew_list_output.split('\n') if x != ""]
            return len(self.specific_packages) == len(brew_list_installed_pkgs)
        else:
            return True


def package_group_for_my_distro():
    """Factory that returns a class representing the distro currently being used."""
    if sys.platform == "darwin":
        return MacOSBrewPackageGroup
    elif sys.platform in ["linux", "linux2", "linux3", ]:
        LINUX_DISTROS = {
            "linuxmint": UbuntuPackageGroup,
            "redhat": RedHatPackageGroup,
            "fedora": FedoraPackageGroup,
            "centos": CentOSPackageGroup,
            "manjarolinux": ArchPackageGroup,
            "arch": ArchPackageGroup,
            "debian": DebianPackageGroup,
            "debian-squeeze": DebianSqueezePackageGroup,
            "debian-squeeze-lts": DebianSqueezeLTSPackageGroup,
            "debian-wheezy": DebianWheezyPackageGroup,
            "debian-jessie": DebianJessiePackageGroup,
            "debian-stretch": DebianStretchPackageGroup,
            "debian-sid": DebianSidPackageGroup,
            "ubuntu": UbuntuPackageGroup,
            "ubuntu-precise": UbuntuPrecisePackageGroup,
            "ubuntu-trusty": UbuntuTrustyPackageGroup,
            "ubuntu-vivid": UbuntuVividPackageGroup,
            "ubuntu-wily": UbuntuWilyPackageGroup,
            "ubuntu-xenial": UbuntuXenialPackageGroup,
        }

        this_distro = utils.lsb_release().lower()

        if this_distro in ["debian", "ubuntu",]:
            return LINUX_DISTROS[
                "{0}-{1}".format(
                    this_distro, utils.lsb_release_codename().lower()
                )
            ]
        else:
            if this_distro in LINUX_DISTROS:
                return LINUX_DISTROS[this_distro]
            else:
                raise exceptions.UnsupportedPlatform(this_distro)
    else:
        raise exceptions.UnsupportedPlatform(sys.platform)
